"""
Utility methods for handling input and output files from blocks and their
respective metadata in `data.json`.
"""
from typing import Any, Tuple, Optional, Union
from pathlib import Path
from geojson import Feature

DATA_PATH = "up42.data_path"


def set_data_path(feature: Feature, data_path: Any) -> Feature:
    """Set path to file in a output feature. Key for data path is always `up42.data_path`.

    Parameters:
        feature: The output feature.
        data_path: Relative path to the output file from the main working folder
            (i.e. `output.tif` is `/tmp/output/output.tif`).

    Returns:
        Output feature with the data path set.
    """
    feature["properties"][DATA_PATH] = data_path
    return feature


def get_data_path(feature: Feature) -> Any:
    """Get path to file in a input feature. Key for data path is always `up42.data_path`.

    Parameters:
        feature: The input feature.

    Returns:
        Relative path to the input file from the main working folder
        (i.e. `input.tif` is `/tmp/input/input.tif`).
    """
    return feature["properties"][DATA_PATH]


def get_output_filename_and_path(
    input_file_name: Union[str, Path],
    postfix: str = "",
    mkdir: bool = True,
    out_path_dir: Path = Path("/tmp/output/"),
    out_file_extension: str = "",
) -> Tuple[str, Path]:
    """
    Utility to generate both output file name and also output file path. Will
    create parent(s) directory(ies) of output file by default.

    Parameters:
        input_file_name :
            File name you can get with `get_data_path` method.
        postfix :
            Additional string to add to the end of the file name before the file suffix or
            format. By default it adds "_" plus postfix.
        mkdir :
            If set to true (default) will create parents of output file path.
        out_path_dir :
            Output path parents. By defaults /tmp/output.
        out_file_extension:
            Output file extension or suffix. Will use input file extension if not explicitly
            set (i.e. .tiff).

    Returns:
        Tuple with str of output file name and output file path that can be passed
        to process to save as output file.
    """
    input_file_name = Path(input_file_name)  # type: ignore
    if out_file_extension == "":
        out_file_extension = input_file_name.suffix
    if postfix == "":
        output_file_name = input_file_name.parent / Path(
            input_file_name.stem + out_file_extension
        )
    else:
        output_file_name = input_file_name.parent / Path(
            input_file_name.stem + "_" + postfix + out_file_extension
        )
    output_file_path = out_path_dir / output_file_name
    if mkdir:
        output_file_path.parent.mkdir(parents=True, exist_ok=True)
    return str(output_file_name), output_file_path


def get_in_out_feature_names_and_paths(
    in_feature: Feature,
    postfix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> Tuple[str, str, Path, Path]:
    """
    Utility to generate the input and output feature names and file paths. Will create
    parent directory(ies) of output file by default. Optionally augments the output filename
    with a postfix.

    Parameters:
        in_feature : A Feature of a GeoJSON FeatureCollection describing all input datasets
        postfix : Additional string to add to the end of the file name before
            the file suffix, adds "_" plus postfix. E.g. "lvl3".
        suffix: Suffix replacement, requires dot e.g. ".tif".

    Returns:
        Tuple with str of in- & output feature names and in- & output file paths.
    """
    in_feature_name = in_feature["properties"][DATA_PATH]
    in_feature_path = Path("/tmp/input") / in_feature_name

    if postfix is None:
        out_feature_name = Path(in_feature_name)
    else:
        in_feature_name = Path(in_feature_name)
        out_feature_name = in_feature_name.with_name(
            in_feature_name.stem + "_" + postfix + in_feature_name.suffix
        )
    if suffix is not None:
        out_feature_name = out_feature_name.with_suffix(suffix)

    out_feature_path = Path("/tmp/output") / out_feature_name
    out_feature_path.parent.mkdir(parents=True, exist_ok=True)

    return (
        str(in_feature_name),
        str(out_feature_name),
        in_feature_path,
        out_feature_path,
    )



"""
Utilities for handling [SpatioTemporal Asset Catalog (STAC)](https://stacspec.org/) queries.
"""

from __future__ import annotations
from dataclasses import dataclass, field

import json
from decimal import Decimal
from typing import Union, Tuple, List, Iterable, Optional, Callable, Any

import shapely.geometry

from geojson.geometry import Geometry
import ciso8601

from .logging import get_logger
from .exceptions import UP42Error, SupportedErrors

logger = get_logger(__name__)

Number = Union[int, Decimal, float]
Coordinates = Iterable[Number]
BoundingBox = Tuple[Number, Number, Number, Number]


@dataclass
class STACQuery:
    """
    Object representing a STAC query as used by data blocks. Standardized
    arguments are:

    Arguments:
        ids: Unique identifiers to search for.
        bbox: A bounding box type geometry.
        intersects: A geometry to use for searching with intersection operator.
        contains: A geometry to use for searching with contains operator.
        time: Time string in RFC3339 (for points in time) or RFC3339/RFC3339
            (for datetime ranges).
        limit: Maximum number of returned elements.
        time_series: List of time strings in RFC3339 (for points in time) or RFC3339/RFC3339
            (for datetime ranges).

    Additional arguments can be passed to the constructor.
    """

    ids: Optional[List[str]] = field(default=None)
    bbox: Optional[BoundingBox] = field(default=None)
    intersects: Optional[Geometry] = field(default=None)
    contains: Optional[Geometry] = field(default=None)
    time: Optional[str] = field(default=None)
    time_series: Optional[List[str]] = field(default=None)
    limit: Optional[int] = field(default=1)

    def __new__(cls, **kwargs):
        """
        Inspired by https://stackoverflow.com/a/63291704/14236301
        Adds attributes to the data class if they are passed in
        input dict in __init__, even if they are not explicitly defined as
        field attributes.
        """
        try:
            initializer = cls.__initializer
        except AttributeError:
            # Store the original init on the class in a different place
            cls.__initializer = initializer = cls.__init__
            # replace init with something harmless
            cls.__init__ = lambda *a, **k: None

        # code from adapted from Arne
        added_args = {}
        for name in list(kwargs.keys()):
            if name not in cls.__annotations__:
                added_args[name] = kwargs.pop(name)

        ret = object.__new__(cls)
        initializer(ret, **kwargs)
        # ... and add the new ones by hand
        for new_name, new_val in added_args.items():
            setattr(ret, new_name, new_val)

        return ret

    def __post_init__(self):
        """
        Post processes the passed arguments to make sure they are valid,
        i.e. only one of "bbox", "intersects" or "contains" is passed.
        """
        if (
            self.bbox
            and self.intersects
            or self.intersects
            and self.contains
            or self.contains
            and self.bbox
        ):
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                """Only one of the following query parameters is
                allowed at a query at any given time:
                * bbox
                * intersects
                * contains.""",
            )

        if self.time and self.time_series:
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                """Only one of the following query parameters is
                allowed at a query at any given time:
                * time
                * time_series
                """,
            )

        if not self.validate_datetime_str(self.time):
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                """Time string could not be validated.
                It must be one of the following formats:
                <RFC3339> (for points in time)
                <RFC3339>/<RFC3339> (for datetime ranges)""",
            )

        if self.limit is not None and self.limit < 1:
            logger.warning(
                "WARNING: limit parameter cannot be < 1, and has been automatically set to 1"
            )
            self.limit = 1

        if self.time_series is not None:
            for datestr in self.time_series:
                if not self.validate_datetime_str(datestr):
                    raise UP42Error(
                        SupportedErrors.WRONG_INPUT_ERROR,
                        """Time string from time_series could not be validated.
                        It must be one of the following formats:
                        <RFC3339> (for points in time)
                        <RFC3339>/<RFC3339> (for datetime ranges)""",
                    )

    def bounds(self) -> BoundingBox:
        """
        Get the bounding box of the query AOI(s) as a BoundingBox object.

        Returns:
            A bounding box object.
        """
        if self.bbox:
            return self.bbox
        elif self.intersects:
            return shapely.geometry.shape(self.intersects).bounds
        elif self.contains:
            return shapely.geometry.shape(self.contains).bounds
        else:
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                """STACQuery does not contain any of the following query parameters:
                * bbox
                * intersects
                * contains.""",
            )

    def geometry(self) -> Geometry:
        """
        Get the geometry of the query AOI(s) as a GeoJson Polygon object.

        Returns:
            A GeoJson Polygon object
        """
        if self.bbox:
            return json.loads(
                json.dumps(shapely.geometry.mapping(shapely.geometry.box(*self.bbox)))
            )
        elif self.intersects:
            return self.intersects
        elif self.contains:
            return self.contains
        else:
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                """STACQuery does not contain any of the following query parameters:
                * bbox
                * intersects
                * contains.""",
            )

    def get_param_if_exists(self, key: str, default_value: Any = None) -> Any:
        """
        Get a query parameter if exists in he STAC query object. Returns
        default_value if not in the query.

        Arguments:
            key: An identifier of the parameter
            default_value: The default return value if the parameter does not exist.
        """
        return getattr(self, key, default_value)

    def set_param_if_not_exists(self, key: str, value: Any):
        """
        Set a query parameter if it does not exist in the STAC query object.

        Arguments:
            key: An identifier of the parameter
            value: The parameter value to set.
        """
        if self.get_param_if_exists(key) is None:
            setattr(self, key, value)

    @classmethod
    def from_json(cls, json_data: str) -> STACQuery:
        """
        Get a STACQuery from a json string representation of a query.

        Arguments:
            json_data: String representation of the query.

        Returns:
            A STACQuery object/
        """
        return STACQuery.from_dict(json.loads(json_data))

    @classmethod
    def from_dict(
        cls, dict_data: dict, validator: Callable[[dict], bool] = lambda x: True
    ) -> STACQuery:
        """
        Get a STACQuery from a dict representation of a query.

        Arguments:
            dict_data: Dictionary representation of the query.
            validator: Function used to validate the query.

        Returns:
            A STACQuery object.
        """

        if not validator(dict_data):
            raise UP42Error(
                SupportedErrors.WRONG_INPUT_ERROR,
                """Input Query did not pass validation. Please refer
                to the official block documentation or block specification.""",
            )

        return STACQuery(**dict_data)

    @staticmethod
    def validate_datetime_str(string: Optional[str]) -> bool:
        """
        Validate a datetime string.

        Arguments:
            string: A datetime string representation.

        Returns:
            If datetime string is valid.
        """
        try:
            if string is None:
                return True
            elif len(str(string).split("/")) == 2:
                ciso8601.parse_datetime(str(string).split("/")[0])
                ciso8601.parse_datetime(str(string).split("/")[1])
            else:
                ciso8601.parse_datetime(str(string))
        except ValueError:
            return False
        return True



"""
Utility to create logger instances.
"""

import logging

LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"


def get_logger(name: str, level=logging.DEBUG) -> logging.Logger:
    """
    Instantiate a logger with a given level and name.

    Example:
        ```python
        logger = blockutils.logging.get_logger(__name__)
        # __name__ is the name of the file where the logger is instantiated.
        ```

    Arguments:
        name: A name for the logger - is included in all the logging messages.
        level: A logging level (i.e. logging.DEBUG).

    Returns:
        A logger instance.
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter(LOG_FORMAT)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger



import sys
import importlib
import pkgutil


def import_submodules(package_name):
    """Import all submodules of a module, recursively

    From https://stackoverflow.com/a/25083161
    """
    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + "." + name)
        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)
    }


__all__ = import_submodules(__name__).keys()



"""
Utilities for handling geometry operations.
"""
import copy
import math
from typing import Generator, List

import geojson
import mercantile
import numpy as np
import rasterio as rio
from area import area
from shapely.geometry import box, mapping, shape
from shapely.validation import explain_validity

from .exceptions import SupportedErrors, UP42Error
from .logging import get_logger
from .stac import BoundingBox, STACQuery

logger = get_logger(__name__)


def filter_tiles_intersect_with_geometry(
    tiles: List[mercantile.Tile], geometry: geojson.geometry.Geometry
) -> Generator[mercantile.Tile, None, None]:
    """
    Given a list of WMTS tiles, this function filters out all tiles which
    don't intersect with the supplied geojson geometry.

    Arguments:
        tiles: List of tiles to filter from.
        geometry: Geometry to use for filtering.

    Returns:
        Tile generator with only tiles that intersect geometry.
    """
    geometry = shape(geometry)
    for tile in tiles:
        tile_bbox = box(*mercantile.bounds(tile)._asdict().values())
        if geometry.intersects(tile_bbox) and not geometry.touches(tile_bbox):
            yield tile


def intersect_geojson_polygons(
    geom1: geojson.geometry.Geometry, geom2: geojson.geometry.Geometry
) -> geojson.geometry.Geometry:
    """
    Convenience method for intersecting two geojson polygons.

    Arguments:
        geom1: A geometry.
        geom2: A different geometry to intersect with.

    Returns:
        Geometry of the intersection.
    """
    geometry1 = shape(geom1)
    geometry2 = shape(geom2)
    intersection = geometry1.intersection(geometry2)

    return mapping(intersection)


def get_query_bbox(query: STACQuery) -> BoundingBox:
    """
    From a STACQuery object return the bounding box of the query.

    Arguments:
        query: A STACQuery object.

    Returns:
        A bounding box object of the query.

    Important:
        If the geometry type is Point or a very small area, a very small buffer
        will be computed and the bounding box of the buffered geometry returned.
    """

    bbox: BoundingBox = query.bounds()

    poly = box(*bbox)
    if poly.area == 0.0:
        # If the geometry in the query was just a point (rather than a box),
        # then make a very small polygon around it so we can derive a
        # bounding box for the OABM API.
        bbox = poly.buffer(0.0001).bounds

    return bbox


def aoi_size_check(
    query: STACQuery, min_sqkm: float = None, max_sqkm: float = None
) -> bool:
    """
    Checks if the query geometry is within set AOI size bounds.

    Arguments:
        query: A STACQuery object.
        min_sqkm: The minimum size in square kilometers.
        max_sqkm: The maximum size in square kilometers.

    Returns:
        Is the query within AOI size bounds.
    """
    valid = True
    area_sqkm = area(query.geometry()) / (10 ** 6)

    if min_sqkm is not None:
        if area_sqkm < min_sqkm:
            valid = False

    if max_sqkm is not None:
        if area_sqkm > max_sqkm:
            valid = False

    return valid


# pylint: disable=chained-comparison
def get_utm_zone_epsg(lon: float, lat: float) -> int:
    """
    Calculates the suitable UTM crs epsg code for an input geometry point.

    Adapted from https://stackoverflow.com/a/55781862/6400555.

    Arguments:
        lon : Longitude of point in degrees (WGS84)
        lat : Latitude of point in degrees (WGS84)

    Returns:
        EPSG code of suitable UTM zone for given (lon, lat). i.e. 32658
    """
    zone_number = int((math.floor((lon + 180) / 6) % 60) + 1)

    # Special zones for Norway
    if lat >= 56.0 and lat < 64.0 and lon >= 3.0 and lon < 12.0:
        zone_number = 32
    # Special zones for Svalbard
    elif lat >= 72.0 and lat < 84.0:
        if lon >= 0.0 and lon < 9.0:
            zone_number = 31
        elif lon >= 9.0 and lon < 21.0:
            zone_number = 33
        elif lon >= 21.0 and lon < 33.0:
            zone_number = 35
        elif lon >= 33.0 and lon < 42.0:
            zone_number = 37

    if lat > 0:
        utm_code = zone_number + 32600
    else:
        utm_code = zone_number + 32700

    return utm_code


def count_vertices(geojson_geom: dict) -> int:
    """Method for counting number of vertices in any kind of geojson geometry\n
    Works for all geometry types:\n
    1. (Multi)Point\n
    2. (Multi)Linestring\n
    3. (Multi)Polygon\n

    Args:
        geojson_geom (dict): e.g. `{"type": "Polygon", "coordinates": [[...]]}`

    Returns:
        int: number of vertices in a geometry
    """
    total_vertices = 0

    coords = geojson_geom["coordinates"]
    if "Multi" in geojson_geom["type"]:
        for c in coords:
            num_vertices = len(c[0])
            total_vertices += num_vertices
    else:
        total_vertices = len(coords[0])

    assert total_vertices > 0, "no vertices were in geometry, check again!"

    return total_vertices


def isclose(
    transform1: rio.Affine,
    transform2: rio.Affine,
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> bool:
    """
    Operator to compare two Affine transformations. Returns true if equal under
    tolerances.

    The tolerance values are positive, typically very small numbers.
    The relative difference (rtol * abs(b)) and the absolute difference
    atol are added together to compare against the absolute difference
    between a and b.

    Arguments:
        transform1: A rasterio transform object.
        transform2: Another rasterio transform object.
        rtol: Relative tolerance for comparison.
        atol: Absolute tolerance for comparison.

    Returns:
        If the two transforms are equal within the given tolerances.

    """
    transform1_ar = np.array(transform1)
    transform2_ar = np.array(transform2)
    return np.allclose(transform1_ar, transform2_ar, rtol=rtol, atol=atol)


def meta_is_equal(meta1: dict, meta2: dict, rtol=1e-05, atol=1e-08) -> bool:
    """
    Operator for comparing two metadata tags provided by a rasterio dataset.

    Compares affine transformations with isclose method.

    Arguments:
        meta1: A meta object.
        meta2: Another meta object.
        rtol: Relative tolerance for comparison.
        atol: Absolute tolerance for comparison.

    Returns:
        If the two meta objects are equal within the given tolerances.

    """
    meta1_transform = meta1.get("transform")
    meta2_transform = meta2.get("transform")

    meta1 = copy.deepcopy(meta1)
    meta2 = copy.deepcopy(meta2)

    del meta1["transform"]
    del meta2["transform"]

    return isclose(meta1_transform, meta2_transform, rtol, atol) and meta1 == meta2


def check_validity(query_geom: dict):
    """checks for validity of the input geometry supplied by user

    Args:
        query_geom (dict): geojson like dict

    Raises:
        UP42Error: raises up42 INPUT_PARAMETERS_ERROR with validity explanation
    """
    shapely_geom = shape(query_geom)

    if not shapely_geom.is_valid:
        explanation = explain_validity(shapely_geom)
        raise UP42Error(
            SupportedErrors.INPUT_PARAMETERS_ERROR,
            f"Input geometry is not valid due to following reason: {explanation}",
        )
